(* main.ml *)
(* driver for an Elkhound parser written in OCaml *)

open Lexerint      (* tLexerInterface *)
open Lexer         (* tToken, readToken *)
open Parsetables   (* tParseTables *)
open Useract       (* tUserActions *)
open Glr           (* tGLR, makeGLR, glrParse *)

open Ptreeact      (* tParseTreeLexer, makeParseTreeActions *)
open Ptreenode     (* tPTreeNode, printTree *)
open Util

class tLexer lexbuf =
object (self)
  inherit tLexerInterface

  val lexbuf:Lexing.lexbuf = lexbuf 

  method getToken() : unit =
  begin
    (* read from stdin *)
    let t:tToken = (readToken lexbuf) in
    (* break the tToken apart into a kind and an sval; perhaps
     * this belongs in lexer.mll too? *)
    tokType <- (tokenKind t);
    match t with
    | Lexer.SMALL_IDENT(s) 
    | Lexer.BIG_IDENT(s) -> sval <- (Obj.repr s)
    | _ -> ()
  end

  method tokenDesc() : string =
  begin
    let kindDesc:string = (self#tokenKindDesc tokType) in
    kindDesc
  end

  method tokenKindDesc (kind:int) : string =
  begin
    (Lexer.tokenKindDesc kind)
  end
end

let split name =
  try 
    let base = Filename.chop_extension name in
    let ext = String.sub name ((String.length base)+1)
      ((String.length name) - ((String.length base)+1))
    in 
    base,ext
  with _ -> name,""

(* --------------------- main -------------------- *)
let main() : unit =
begin
  let do_parse_tree = ref false in 
  let files = ref [] in 

  (* create the lexer *)
  let argDescr = [
    "--tree", Arg.Set do_parse_tree, "\tjust print the parse tree" ;
  ] in
  let give_help () = Arg.usage argDescr "" ; exit 1 in 
  let handleArg str = begin
    let f,e = split (String.uppercase str) in
    match e with
    | "ML"
    | "MLI" -> files := str :: !files 
    | _ -> give_help () 
  end in
  Arg.parse argDescr handleArg "" ; 

  let (tables:tParseTables), (actions:tUserActions) =
    (Oparse.oparseParseTables, Oparse.oparseUserActions)
  in

  List.iter (fun filename -> 
    try 
      let inchan = open_in filename in 
      let lexbuf = lex_init filename inchan in 
      let lex:tLexerInterface = ((new tLexer lexbuf) :> tLexerInterface) in
      (* prime the lexer: get first token *)
      (lex#getToken());
      (* get parse tables and user actions, depending on which
       * grammar the user wants to use *)
      (* substitute tree-building actions? *)
      let (lex:tLexerInterface), (actions:tUserActions) =
        if (!do_parse_tree) then (
          ((new tParseTreeLexer lex actions),      (* tree lexer *)
          (makeParseTreeActions actions tables))  (* tree actions *)
        )
        else (lex, actions)                           (* unchanged *)
      in 
      (* parse the input *)
      let (sval:tSemanticValue) = 
          (* use GLR *)
          let glr:tGLR = (makeGLR tables actions) in
          let treeTop: tSemanticValue ref = ref cNULL_SVAL in
          if (not (glrParse glr lex treeTop)) then (
            (parse_error "GLR parse error")
          );
          !treeTop
      in
      close_in inchan ; 
      (* interpret the result *)
      if (!do_parse_tree) then begin
        (* it's a tree *)
        let t:tPTreeNode = ((Obj.obj sval) : tPTreeNode) in
        (printTree t stdout true(*expand*));
        (flush stdout);
      end else begin
        let (tdll:Otype.type_def list list) = 
          ((Obj.obj sval) : Otype.type_def list list) in
        Printf.printf "%s: %d types in %d groups.\n" filename 
          (List.length (List.flatten tdll))
          (List.length tdll) ;

        (* let d = Otype_print.d_type_def_list () tdl in
        Pretty.fprint stdout 70 d ;   
        Printf.printf "\n" ;  
        *)

        let d = Pp.emit_pp tdll in
        let f,e = split filename in
        let out_filename = f ^ "_print." ^ e in 
        let outchan = open_out out_filename in 
        Printf.fprintf outchan "open Pretty\nopen %s\n\n"
          (String.capitalize (String.lowercase f)) ;
        Pretty.fprint outchan 80 d ;

        let d = Pp.emit_v f tdll in
        Pretty.fprint outchan 80 d ;

        close_out outchan 
      end 
    with e -> 
      Printf.printf "%s: %s\n" filename 
        (Printexc.to_string e) 
  ) !files
end
;;

Printexc.catch main()
;;
